#include <CapacitiveSensor.h>
#include <OneWire.h>
#include <DallasTemperature.h>

#define bluePin 2                            //LEDs
#define greenPin 3
#define yellowPin 4
#define redPin 5
#define whitePin 6

#define tempPin 10
#define voltPin A2
#define pumpPin A3
#define levelPin A1
//#define levelPin2 A0        NOT USED
#define levelPowerPin 11
#define BTResetPin 7

#define pumpRecheckInterval 80  //in polls
#define pollDelay 250
#define minVolts 8.0
#define minLevelReading 14500
#define minTemp 32.5            //Degrees F
#define BTTimeout 120000
#define minStartVoltsMult 1.4
#define tempLevelPollsPerCheck 20

long lastPoll, pumpOnTime, lastPumpOnTime, pumpOnTotal, lastCommandTime ;
boolean bBlue, bGreen, bYellow, bRed, bWhite, bPump ;
float lastVoltage, minVoltage, maxVoltage, lastTempInt, maxTempInt, minTempInt, lastTempExt, maxTempExt, minTempExt ;
long lastLevel, minLevel, maxLevel, pollCount ;

OneWire oneWire(tempPin) ;
DallasTemperature tempSensors(&oneWire);
CapacitiveSensor levelSensor = CapacitiveSensor(levelPowerPin, levelPin) ;

DeviceAddress insideThermometer, outsideThermometer;

void setup() {
  pinMode(bluePin, OUTPUT) ;
  pinMode(greenPin, OUTPUT) ;
  pinMode(yellowPin, OUTPUT) ;
  pinMode(redPin, OUTPUT) ;
  pinMode(whitePin, OUTPUT) ;
  pinMode(voltPin, INPUT) ;
  pinMode(pumpPin, OUTPUT) ;
  pinMode(levelPowerPin, OUTPUT) ;
  pinMode(BTResetPin, OUTPUT) ;

  Serial.begin(9600) ;
  tempSensors.begin() ;

  if (!tempSensors.getAddress(insideThermometer, 0)) Serial.println("Unable to find address for internal temp sensor"); 
  if (!tempSensors.getAddress(outsideThermometer, 1)) Serial.println("Unable to find address for external temp sensor"); 
  
  bBlue = false ;
  bGreen = false ;
  bYellow = false ;
  bRed = false ;
  bWhite = false ;
  bPump = false ;
  lastVoltage = 0.0 ;
  minVoltage = 50 ;
  maxVoltage = -1 ;
  lastTempInt = 0.0 ;
  minTempInt = 200.0 ;
  maxTempInt = -100.0 ;
  lastTempExt = 0.0 ;
  minTempExt = 200.0 ;
  maxTempExt = -100.0 ;
  lastLevel = 0 ;
  minLevel = 50000 ;
  maxLevel = -1 ;
  pollCount = 0 ;
  
  pumpOnTime = 0 ;
  pumpOnTotal = 0 ;
  lastPumpOnTime = 0 ;
  lastCommandTime = -55000 ;

  digitalWrite(BTResetPin, HIGH) ;
  digitalWrite(pumpPin, LOW) ;
  update() ;
  readLevelTemp() ;
  readVolts() ;
  lastPoll = millis() ;

  Serial.println("Initialized") ;

}

void loop() {
  char tempChar ;
  String serialInput = "" ;
  float recheckVolts ;
  int reading ;

  if(lastPoll + pollDelay <= millis()) {              //Check to poll
    pollCount++ ;
    lastPoll = millis() ;
    readVolts() ;
    
    if(pollCount % tempLevelPollsPerCheck == 0)      //Check 
      readLevelTemp() ;

    evaluate() ;
  }   
  
  if(pollCount % tempLevelPollsPerCheck < 3)        //750ms blinking heartbeat to show the program is running
    bGreen = true ;
  else
    bGreen = false ;

  update() ;

  while(Serial.available()){
    tempChar = Serial.read() ;
    serialInput = serialInput + tempChar ;
  }

  if(serialInput != "")
    if(serialInput.indexOf("read") != -1) {
      lastCommandTime = millis() ;
      readout() ;
    }

  if(lastCommandTime + BTTimeout < millis()) {                  //If you haven't heard a request on BT recently, reset the module (brownout protection)
    lastCommandTime = millis() ;
    digitalWrite(BTResetPin, LOW) ;
    delay(500) ;
    digitalWrite(BTResetPin, HIGH) ;
  }
  
  delay(5) ;
}


void evaluate() {                                      //Checking
  if(lastTempInt < minTemp || lastTempExt < minTemp)
    bYellow = true ;
  else
    bYellow = false ;
  if(lastLevel < minLevelReading)
    bRed = true ;
  else
    bRed = false ;
  if(lastVoltage < minVolts)
    bWhite = true ;
  else
    bWhite = false ;

  if(lastTempInt > minTemp && lastTempExt > minTemp && lastLevel > minLevelReading && lastVoltage >= minVolts){
    if(bPump) {
      pumpOnTotal += millis() - lastPumpOnTime ;
      pumpOnTime += millis() - lastPumpOnTime ;
      lastPumpOnTime = millis() ;
    }
    else if(pollCount % pumpRecheckInterval == 0){                    //Only turn the pump on if it's allowed
      if(lastVoltage >= minVolts * minStartVoltsMult){
        bPump = true ;
        lastPumpOnTime = millis() ;
      }
      else
        bWhite = true ;
    }
  }
  else {
    if(bPump){
      pumpOnTotal += millis() - lastPumpOnTime ;
    }
    bPump = false ;
    pumpOnTime = 0 ;
  }
  

  bBlue = bPump ;
}

void readout() {                                                  //The reply over bluetooth to "read"
  Serial.println("***Report***") ;
  Serial.print("Time: ") ;
  Serial.print((millis() / 3600000)) ;
  Serial.print("h ") ;
  Serial.print(((millis() % 3600000) / 60000)) ;
  Serial.print("m ") ;
  Serial.print(((millis() % 60000) / 1000)) ;
  Serial.println("s") ;
  Serial.print("Internal Temp: " ) ;
  Serial.print(lastTempInt) ;
  Serial.print("F, min: ") ;
  Serial.print(minTempInt) ;
  Serial.print(" max: ") ;
  Serial.println(maxTempInt) ;
  Serial.print("Water Temp: ") ;
  Serial.print(lastTempExt) ;
  Serial.print("F, min: ") ;
  Serial.print(minTempExt) ;
  Serial.print(" max: ") ;
  Serial.println(maxTempExt) ;
  Serial.print("Solar Voltage: ") ;
  Serial.print(lastVoltage) ;
  Serial.print("V,  min: ") ;
  Serial.print(minVoltage) ;
  Serial.print(" max: ") ;
  Serial.println(maxVoltage) ;
  Serial.print("Water Level: ") ;
  Serial.print(lastLevel) ;
  Serial.print(", min: ") ;
  Serial.print(minLevel) ;
  Serial.print(" max: ") ;
  Serial.println(maxLevel) ;
  Serial.print("Pump is ") ;
  if(bPump) {
    Serial.print("on for ") ;
    Serial.print((pumpOnTime / 3600000)) ;
    Serial.print("h ") ;
    Serial.print(((pumpOnTime % 3600000) / 60000)) ;
    Serial.print("m ") ;
    Serial.print(((pumpOnTime % 60000) / 1000)) ;
    Serial.println("s") ;
  }
  else {
    Serial.print("off because: ") ;
    if(bYellow)
      Serial.println("temperature is too low") ;
    else if(bRed)
      Serial.println("water level is too low") ;
    else if(bWhite)
      Serial.println("solar voltage is too low") ;
    else
      Serial.println("pump restart delay") ;
  }
  Serial.print("Total pump uptime: ") ;
  Serial.print((pumpOnTotal / 3600000)) ;
  Serial.print("h ") ;
  Serial.print(((pumpOnTotal % 3600000) / 60000)) ;
  Serial.print("m ") ;
  Serial.print(((pumpOnTotal % 60000) / 1000)) ;
  Serial.println("s") ;
}

void update() {                              //Write LEDs
  digitalWrite(bluePin, !bBlue) ;
  digitalWrite(greenPin, !bGreen) ;
  digitalWrite(yellowPin, !bYellow) ;
  digitalWrite(redPin, !bRed) ;
  digitalWrite(whitePin, !bWhite) ;
  digitalWrite(pumpPin, bPump) ;
}

void readVolts() {
  int reading = 0 ;
  reading = analogRead(voltPin) ;
  lastVoltage = (float)(reading * 5 * 4.29 / 1023) ;

  if(lastVoltage < minVoltage)
    minVoltage = lastVoltage ;
  if(lastVoltage > maxVoltage)
    maxVoltage = lastVoltage ;
}

void readLevelTemp() {
  tempSensors.requestTemperatures() ;
  lastTempInt = tempSensors.getTempF(insideThermometer) ;
  lastTempExt = tempSensors.getTempF(outsideThermometer) ;
  
  lastLevel = levelSensor.capacitiveSensorRaw(100) ;
  
  if(lastTempInt < minTempInt)
    minTempInt = lastTempInt ;
  if(lastTempInt > maxTempInt)
    maxTempInt = lastTempInt ;
  if(lastTempExt < minTempExt)
    minTempExt = lastTempExt ;
  if(lastTempExt > maxTempExt)
    maxTempExt = lastTempExt ;
  if(lastLevel < minLevel)
    minLevel = lastLevel ;
  if(lastLevel > maxLevel)
    maxLevel = lastLevel ;
}


